VERSION 5.00
Begin VB.UserControl Anisim 
   ClientHeight    =   3600
   ClientLeft      =   0
   ClientTop       =   0
   ClientWidth     =   4800
   ScaleHeight     =   3600
   ScaleWidth      =   4800
   Begin VB.PictureBox pb 
      AutoRedraw      =   -1  'True
      BackColor       =   &H00000000&
      Height          =   2055
      Left            =   1080
      ScaleHeight     =   133
      ScaleMode       =   3  'Pixel
      ScaleWidth      =   165
      TabIndex        =   0
      Top             =   480
      Width           =   2535
   End
   Begin VB.Timer Timer1 
      Left            =   360
      Top             =   480
   End
End
Attribute VB_Name = "Anisim"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
Option Explicit

Public Nodes As New Collection
Public Edges As New Collection

Private Selected As Boolean
Private SelNode As CNode

Private m_step As Long
Private m_repelNode As Long
Private m_repelBoundary As Long
Private m_friction As Double

Const NODE_SIZE = 8

Private Sub pb_MouseDown(Button As Integer, Shift As Integer, x As Single, y As Single)
    Dim n As CNode
        
    'Debug.Print "MouseDown", x, y
    
    For Each n In Nodes
        Dim dx As Double
        Dim dy As Double
        Dim dis As Double
        
        dx = x - n.x
        dy = y - n.y
        
        dis = Sqr(dx * dx + dy * dy)
        
        If dis < NODE_SIZE Then
            n.fixed = True
            Set SelNode = n
            Selected = True
            Redraw
            Exit Sub
        End If
    Next
End Sub

Private Sub pb_MouseMove(Button As Integer, Shift As Integer, x As Single, y As Single)
    If Selected Then

        'Debug.Print "MouseMove", x, y
        
        SelNode.x = x
        SelNode.y = y
        
        Redraw
    End If
End Sub

Private Sub pb_MouseUp(Button As Integer, Shift As Integer, x As Single, y As Single)
    If Selected Then
        'Debug.Print "MouseUp", x, y
        SelNode.fixed = False
        Selected = False
        Redraw
    End If
End Sub

Private Sub Timer1_Timer()
    pb.Cls
    Relax
    m_step = m_step + 1
    NextFrame
End Sub

Private Sub UserControl_Resize()
    pb.Top = 0
    pb.Left = 0
    pb.Width = UserControl.Width
    pb.Height = UserControl.Height
End Sub

Public Sub Redraw()
    pb.Cls
    NextFrame
End Sub

Private Sub NextFrame()
    Dim n As CNode
    Dim e As CEdge
    
    For Each e In Edges
        e.Draw pb
    Next
    
    For Each n In Nodes
        n.Draw pb, NODE_SIZE
    Next
    
    'PrintAt 10, 10, "Step = " & m_step
    
    'Dim i
    'For i = 1 To 16
    '    pb.CurrentX = 20
    '    pb.CurrentY = i * 20 + 20
    '    pb.ForeColor = QBColor(11)
    '    pb.Print i Mod 7
    '    pb.DrawStyle = i Mod 7
    '    pb.Line (50, i * 20 + 20)-(150, i * 20 + 20), QBColor(i - 1)
    'Next
End Sub

Private Sub PrintAt(x As Integer, y As Integer, msg As String)
    pb.ForeColor = QBColor(12)
    pb.CurrentX = x
    pb.CurrentY = y
    pb.Print msg
End Sub

Private Sub Relax()
    ' relocate nodes and edges based on
    ' the simulated physics of the system
    
    Dim e As CEdge
    Dim n As CNode
    
    Dim x1 As Double
    Dim y1 As Double
    Dim x2 As Double
    Dim y2 As Double
    Dim dx As Double
    Dim dy As Double
    Dim vx As Double
    Dim vy As Double
    Dim dis As Double
    Dim f As Double
    
    Dim colSpark
    colSpark = QBColor(1)
    
    ' adjust inverse-square forces based on control size
    Dim scaling As Double, scaling2 As Double
    scaling = (pb.ScaleWidth / 1000)
    scaling2 = scaling * scaling
        
    For Each e In Edges
        If Not e.enabled Then GoTo Continue
        
        ' enabled edges behave as a spring of
        ' the specified length
        x1 = e.src.x
        y1 = e.src.y
        
        x2 = e.dst.x
        y2 = e.dst.y
        
        vx = x1 - x2
        vy = y1 - y2
        
        dis = Sqr(vx * vx + vy * vy)
        
        ' spring force, spring length adjusted for control size
        f = e.strength * (0.001) * ((e.length * scaling) - dis)
        
        'Debug.Print e.length, Format(f, " 0.0000")
        
        dx = f * (vx / dis)
        dy = f * (vy / dis)
               
        e.src.Push dx, dy
        e.dst.Push -dx, -dy
Continue:
    Next
    
    Dim count As Integer
    Dim i As Integer
    Dim j As Integer
    
    count = Nodes.count
    
    ' all nodes repel each other with an inverse-square relationship
    For i = 1 To count
        For j = i + 1 To count
            Dim n1 As CNode
            Dim n2 As CNode
            
            Set n1 = Nodes(i)
            Set n2 = Nodes(j)
            
            x1 = n1.x
            y1 = n1.y
            
            x2 = n2.x
            y2 = n2.y
            
            vx = x1 - x2
            vy = y1 - y2
            
            'If Abs(vx) < 50 And Abs(vy) < 50 Then
                dis = Sqr(vx * vx + vy * vy)
                If dis > 0 Then
                    f = m_repelNode * 1000 / (dis * dis) * scaling2
                End If
                
                dx = f * (vx / dis)
                dy = f * (vy / dis)
                
                n1.Push dx, dy
                n2.Push -dx, -dy
                
                'pb.Line (x1, y1)-(x2, y2), colSpark
            'End If
           
        Next
    Next
    
    ' the borders repel nodes with an inverse-square relationship
    If m_repelBoundary > 0 Then
        For Each n In Nodes
            
            ' left
            If n.x > 0 Then
                n.vx = n.vx + m_repelBoundary * 1000 / (n.x * n.x) * scaling2
            Else
                n.vx = n.vx + m_repelBoundary * 1000
            End If
            
            ' top
            If n.y > 0 Then
                n.vy = n.vy + m_repelBoundary * 1000 / (n.y * n.y) * scaling2
            Else
                n.vy = n.vy + m_repelBoundary * 1000
            End If
            
            ' right
            If n.x < pb.ScaleWidth Then
                n.vx = n.vx - m_repelBoundary * 1000 / ((pb.ScaleWidth - n.x) * (pb.ScaleWidth - n.x)) * scaling2
            Else
                n.vx = n.vx - m_repelBoundary * 1000
            End If
            
            ' bottom
            If n.y < pb.ScaleHeight Then
                n.vy = n.vy - m_repelBoundary * 1000 / ((pb.ScaleHeight - n.y) * (pb.ScaleHeight - n.y)) * scaling2
            Else
                n.vy = n.vy - m_repelBoundary * 1000
            End If
            
        Next n
    End If

    For Each n In Nodes
        n.Move (m_friction)
    Next
End Sub

Public Sub Start()
    Timer1.Interval = 20
End Sub

Public Sub Pause()
    Timer1.Interval = 0
End Sub

Public Sub TestSet(num As Integer)
    Randomize
    
    Dim xmax As Integer
    Dim ymax As Integer
    
    Dim n As CNode
    Dim e As CEdge
    Dim i As Integer
    Dim j As Integer
    
    ReDim nodearr(1 To num) As New CNode
    
    Set Nodes = New Collection
    Set Edges = New Collection
    
    xmax = pb.ScaleWidth
    ymax = pb.ScaleHeight
    
    For i = 1 To num
        nodearr(i).x = Rnd * xmax
        nodearr(i).y = Rnd * ymax
    
        Nodes.Add nodearr(i)
    Next
    
    For i = 1 To num - 1
        For j = i + 1 To num
            Set e = New CEdge
            
            Set e.src = nodearr(i)
            Set e.dst = nodearr(j)
            e.length = 100
            e.strength = 100
            
            Edges.Add e
        Next
    Next
    
    Redraw
End Sub

Public Sub RandomPlace(n As CNode)
    n.x = Rnd * pb.ScaleWidth
    n.y = Rnd * pb.ScaleHeight
End Sub

Public Sub Place(n As CNode, x As Double, y As Double)
    n.x = x * pb.ScaleWidth
    n.y = y * pb.ScaleHeight
End Sub

Public Property Get ScaleWidth() As Variant
    ScaleWidth = pb.ScaleWidth
End Property

Public Property Get ScaleHeight() As Variant
    ScaleHeight = pb.ScaleHeight
End Property

Public Property Get RepelNode() As Long
    RepelNode = m_repelNode
End Property

Public Property Let RepelNode(ByVal vNewValue As Long)
    If vNewValue >= 0 Then
        m_repelNode = vNewValue
    End If
End Property

Public Property Get RepelBoundary() As Long
    RepelBoundary = m_repelBoundary
End Property

Public Property Let RepelBoundary(ByVal vNewValue As Long)
    If vNewValue >= 0 Then
        m_repelBoundary = vNewValue
    End If
End Property

Public Property Get Friction() As Double
    Friction = m_friction
End Property

Public Property Let Friction(ByVal vNewValue As Double)
    If vNewValue >= 0 And vNewValue <= 1 Then
        m_friction = vNewValue
    End If
End Property


