// $Header: E:/TNG/src/rcs/fsm.cpp 1.2 1997/02/22 14:37:48 dbm Exp $

#include "fsm.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//	bioFSM
//

bioFSM::bioFSM (int is, int ss, int os)
//: is = # of input bits, ss = # of state bits, os = # output bits
{
	assert ((ss + is) <= 32);

    _state = 0;
    _index = 0;
    _output = 0;
    _table = 0;

    _inputSize = is;
    _stateSize = ss;
    _outputSize = os;

    _inputMask = (1 << _inputSize) - 1;
    _stateMask = (1 << _stateSize) - 1;
    _outputMask = (1 << _outputSize) - 1;
}

bioFSM::~bioFSM ()
{
    delete [] _table;
}

void
bioFSM::Init()
{
	// table size = 2^n where n = state bits + input bits
    _tableSize = 1 << (_stateSize + _inputSize); 
    _table = new int[_tableSize]; 
    
    ASSERT (_table);
}

void
bioFSM::Randomize()
{
    for (int i = 0; i < _tableSize; i++)
    {
       _table[i] = rand();
    }
}

void
bioFSM::ReadArray (int* array, int size)
{                                      
	assert (size == _tableSize);
	
    for (int i = 0; i < _tableSize; i++)
    {
    	_table[i] = array[i];
    }
}

#include <stdio.h>

int
bioFSM::Think (int* vector, int size)
{
    int                 i;
    unsigned int		value;

	//printf("Think (before) index=%d state=%d output=%d\n",
	//	_index, _state, _output);

    _index = _state;

    for (i = 0; i < _inputSize; i++)
    {
    	_index = (_index<<1) | vector[i];
    }

    ASSERT (_index >= 0);
    ASSERT (_index < _tableSize);
    
    value = (unsigned int) _table[_index];
    _state = (int) ((value >> _inputSize) & _stateMask);
    _output =  (int) (value & _outputMask);

	//printf("Think (after) index=%d value=%d state=%d output=%d\n",
	//	_index, value, _state, _output);

    return _output;
}

void
bioFSM::GetState (bioBitVector& vector)
{
	vector[0] = (_state >> 0) & 0xff;
	vector[1] = (_state >> 8) & 0xff;
	vector[2] = (_state >> 16) & 0xff;
	vector[3] = (_state >> 24) & 0xff;

	//printf ("GetState %d\n", _state);
}

void
bioFSM::SetState (bioBitVector& vector)
{
	_state = (vector[3] << 24) | (vector[2] << 16) | (vector[1] << 8) | vector[0];

	//printf ("SetState %d\n", _state);
}


