// $Header: E:/TNG/src/rcs/xmatrix.cpp 1.2 1997/02/22 14:38:20 dbm Exp $

#include "xmatrix.h"
  
Point::Point()
{
	x = y = 0;
}
             
Point::Point (int xx, int yy)
{
	x = xx;
	y = yy;
}
            
ostream&
operator<< (ostream& s, const Point& p)
{
	s << "(" << p.x << "," << p.y << ")";
	return s;
}

Point
operator* (Point& p, XMatrix& m)
{
	Point		foo; 

	foo.x = Round (m(0, 0) * p.x + m(1, 0) * p.y + m(2, 0));
	foo.y = Round (m(0, 1) * p.x + m(1, 1) * p.y + m(2, 1));
	
	return foo;
}

ostream&
operator<< (ostream& s, const XMatrix& m)
{            
	for (int i = 0; i < 3; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			s << m(i, j) << " ";
		}         
		             
		s << endl;
	}
	
	return s;
}
    
XMatrix::XMatrix()
{
	Clear();
}

void
XMatrix::Clear()
{
	for (int i = 0; i < 3; i++)
	{
		for (int j = 0; j < 3; j++)
		{
			_val[i][j] = 0;
		}
	}
}

void
XMatrix::Identity()
{
	Clear();
	
	for (int i = 0; i < 3; i++)
	{
		_val[i][i] = 1;
	}
}
                         
double&
XMatrix::operator()(int i, int j)
{               
	return _val[i][j];
}
                            
const double&
XMatrix::operator()(int i, int j) const
{               
	return _val[i][j];
}
                         
XMatrix 
operator* (XMatrix& a, XMatrix& b)
{          
	XMatrix			c;
	
	for (int i = 0; i < 3; i++)
	{
		for (int j = 0; j < 3; j++)
		{
	    	c(i, j) = 0;

			for (int k = 0; k < 3; k++)
			{    
				c(i, j) += a(i, k) * b(k, j);
			}
		}
	}
	
	return c;
}

void
XMatrix::Translate (double tx, double ty)
{
	Identity();

	_val[0][2] = tx;
	_val[1][2] = ty;
}

void
XMatrix::Scale (double sx, double sy)
{
	Identity();

	_val[0][0] = sx;
	_val[1][1] = sy;
}

void
XMatrix::Rotate (double theta)
{
	double		cosTheta = cos(theta);
	double		sinTheta = sin(theta);
	
	Identity();

	_val[0][0] = cosTheta;
	_val[1][0] = -sinTheta;
	
	_val[0][1] = sinTheta;
	_val[1][1] = cosTheta;
}
 
void 
XMatrix::RotateAboutPoint (double theta, double x, double y)
{
	Rotate (theta);
	
	_val[2][0] = (1 - _val[0][0]) * x + _val[0][1] * y;
	_val[2][1] = (1 - _val[0][0]) * y - _val[0][1] * x;
}
 
