//
//  tng.cpp
//
//  Main program for TNG
//  Last Revised: 12 November 2000
//
//                       TNG (Version tnglab.101)
//         A C++ Implementation for the Trade Network Game
//
//        Copyright 2000 David McFadzean and Leigh Tesfatsion.
//                       All Rights Reserved
//
//  Important Disclaimer:
//     This TNG software package is unsupported and provided as-is,
//     without warranty of any kind.  It is released here as freeware
//     under the terms of the Artistic License Agreement residing at
//     http://www.econ.iastate.edu/tesfatsi/alicense.htm
//
//  TNG Manual:
//        D. McFadzean and L. Tesfatsion, "A C++ Platform for the
//        Evolution of Trade Networks," _Computational Economics_
//        14 (1999), pages 109--134, Special Issue on Programming
//        languages.  For information regarding related materials,
//        see http://www.econ.iastate.edu/tesfatsi/vita.htm#ACE
//
//  The TNG consists of a collection of traders that evolve over time.
//  The implementation of the TNG makes use of three principal classes:
//    tngSimulation:  manages the overall simulation;
//    tngPopulation:  manages the evolution of the traders;
//    tngTradeBot:    manages the simulation of a single trader, either
//                    a pure buyer, a pure seller, or a buyer-seller.
//  The TNG program creates a single instance of tngSimulation, which
//  creates one instance of tngPopulation, and tngPopulation creates
//  a collection of tngTradeBot instances.  The TNG is supported by
//  the SimBioSys abstract base classes developed by David McFadzean.
//

#include <stdio.h>
#include "tng.h"        //  Header files for the TNG

int
main (void)
{
  tngSimulation  tng;
  int            g, i;

  // Read the simulation parameters
  char*	iniFile = "tng.ini";
  tng.OpenFile(iniFile);

  tng.Init();

  tng.DumpInput();


// Start the generation cycle loop.

  for (g = 0; g < tng.GMax(); g++)
  {
     if (g > 0)
     {
        tng.EvolveGen();                 // evolution step
     }

     tng.InitGen(g);

     for (i = 0; i < tng.IMax(); i++)    // trade cycle loop
     {
        tng.MatchTraders();
        tng.Trade(i);
     }
                                         // environmental step
     tng.AssessFitness();
	 tng.CalcScores();
	 
	 tng.DumpBots();      // Detailed information is printed out
     tng.Dump();       // for the final generation.
                             
   }

   return 0;
}
