//
// tngSim.cpp (for tnglab.101 -- but not for use with tngshell)
//
// Last Revised: 29 June 1998
//

#include <stdio.h>
#include <time.h>
#include "tngsim.h"

/////////////////////////////////////////////////////////
// population standard deviation utility
float stddevpop(float * val, int count)
{
	if (count <= 0 || val == 0) 
		return 0.0f;

	int i;
	float sum = 0.0;
	for (i = 0; i < count; i++)
	{
		sum += val[i];
	}

	float avg = sum / (float) count;	

	float diff = 0.0;
	float variance = 0.0;
	for (i = 0; i < count; i++)
	{
		diff = val[i] - avg;
		variance += diff * diff;
	}
	variance /= count;	// note "sample standard deviation" would divide by (count - 1) instead

	float stdDev = (float) sqrt(variance);

	return stdDev;
}

tngScores::tngScores()
{
	minScore = +1000.0f;
	avgScore = 0.0f;
	maxScore = -1000.0f;
	totScore = 0.0f;
	stddevScore = 0.0f;
}


//////////////////////////
//
// tngSimulation methods

tngSimulation::tngSimulation(void)
{
	_gmax = 0;
	_imax = 0;

	_seed = 0;
	_mrate = 0.0;
	_fsmStates = 0;
	_fsmMemory = 0;

	_elitePercentPB = 0.0;
	_elitePercentPS = 0.0;
	_elitePercentBS = 0.0;


	_buyerQuota = 0;     
	_sellerQuota = 0;    
	_initExpPayoff = 0.0;
	_refusalPayoff = 0.0;  
	_inactivePayoff = 0.0;
	_experienceGain = 0.0f;

	_bothCoop = 0.0;         
	_bothDefect = 0.0;       
	_temptation = 0.0;       
	_sucker = 0.0;           
	_traderCount = 0;      
	_buyerCount = 0;       
	_sellerCount = 0;      
	_pureBuyerCount = 0;   
	_pureSellerCount = 0;  
	_buysellCount = 0;     
	_trades = 0;           
	_printGen = 0;         
	_printIt = 0; 
	
	_dumpFile = "";
}

tngSimulation::~tngSimulation(void)
{ 
	_traders.RemoveAll();
	_buyers.RemoveAll();
	_sellers.RemoveAll();
}

void
tngSimulation::Init(void)
{    		
	srand((unsigned int) _seed);
	tngTradeBot::SetMemory(_fsmMemory);
	tngTradeBot::SetStates(_fsmStates);

	_trades = 0;


	// initialize agent counts
	_traderCount = _pureBuyerCount + _pureSellerCount + _buysellCount;

	_eliteCountPB  =  (int)(_elitePercentPB * _pureBuyerCount);
	_eliteCountPS  =  (int)(_elitePercentPS * _pureSellerCount);
	_eliteCountBS  =  (int)(_elitePercentBS * _buysellCount);
	
	if (_buysellCount < 0)
	{
		_buysellCount = 0;
	}

	_buyerCount  = _pureBuyerCount  + _buysellCount;
	_sellerCount = _pureSellerCount + _buysellCount;
	
	
	// initialize trade populations
	ASSERT((_pureBuyerCount  >= _eliteCountPB) && (_eliteCountPB >= 0));
	ASSERT((_pureSellerCount >= _eliteCountPS) && (_eliteCountPS >= 0));
	ASSERT((_buysellCount    >= _eliteCountBS) && (_eliteCountBS >= 0));

	if (_pureBuyerCount > 0)
	{
		_pureBuyPopulation.Init();
		_pureBuyPopulation.SetMutationRate(_mrate);
		_pureBuyPopulation.SetElite(_eliteCountPB);
		_pureBuyPopulation.SetSize(_pureBuyerCount);
	}

	if (_pureSellerCount > 0)
	{
		_pureSellPopulation.Init();
		_pureSellPopulation.SetMutationRate(_mrate);
		_pureSellPopulation.SetElite(_eliteCountPS);
		_pureSellPopulation.SetSize(_pureSellerCount);
	}

	if (_buysellCount > 0)
	{
		_buysellPopulation.Init();
		_buysellPopulation.SetMutationRate(_mrate);
		_buysellPopulation.SetElite(_eliteCountBS);
		_buysellPopulation.SetSize(_buysellCount);
	}

	InitTradeBots();
}

void
tngSimulation::InitGen(int g)
{
	_curGen = g;

	DUMP("\nGeneration #%3d\n", _curGen+1);

}

void
tngSimulation::InitTradeBots()
// build trader lists from the populations and 
// set up trade info etc. for all the trade bots. 
{
	int		i;
	tngTradeBot*		bot = 0;

	ASSERT(_traders.IsEmpty());
	ASSERT(_buyers.IsEmpty());
	ASSERT(_sellers.IsEmpty());

	for (i = 0; i < _pureBuyerCount; i++)
	{
		bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _pureSellerCount; i++)
	{
		bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _buysellCount; i++)
	{
		bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _pureBuyerCount; i++)
	{
		bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_buyers.Append(bot);

		bot->Init(this, _traders, _buyerQuota, 0);
	}

	for (i = 0; i < _pureSellerCount; i++)
	{
		bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_sellers.Append(bot);

		bot->Init(this, _traders, 0, _sellerQuota);
	}

	for (i = 0; i < _buysellCount; i++)
	{
		bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_buyers.Append(bot);
		_sellers.Append(bot);

		bot->Init(this, _traders, _buyerQuota, _sellerQuota);
	}
}
	
const char*	
tngSimulation::GetDumpFile()
{
	return _dumpFile;
}

void		
tngSimulation::SetDumpFile(const char* file)
// set the file name for dumping information
// if "" then dump will be to stdout
{
	if (_dumpFile)
		delete _dumpFile;

	strcpy(_dumpFile, file);
}

void		
tngSimulation::DUMP(char* fmt, ...)
// dump formatted string to dump file
{
	char				buffer[1024];
    va_list             argptr;

    va_start (argptr, fmt);
    vsprintf (buffer, fmt, argptr);
    
	
	// append to dump file
	FILE*	fp = _dumpFile != "" ? fopen (_dumpFile, "a+") : stdout;

	if (fp == 0)
	{
		ASSERT(0);
		return;
	}

	fprintf (fp, "%s", buffer);

	if (_dumpFile != "")
		fclose(fp);
}

void
tngSimulation::MatchTraders(void) 
{
	tngTradeBot* 		buyer;
	tngTradeBot* 		seller;
	tngTradeBot* 		trader;
	int					offers;

	for (_buyers.First(); buyer = (tngTradeBot*) _buyers.Next(); )
	{
		buyer->PrepareToMakeOffers();
	}
                   
	for (_sellers.First(); seller = (tngTradeBot*) _sellers.Next(); )
	{
		seller->PrepareToAcceptOffers();
	}
                   
// Tell all buyers to submit offers and all sellers to
// provisionally accept offers until the buyers stop making
// offers.
	while(1)
	{
		offers = 0;

// Tell all buyers to submit their offers...
		for (_buyers.First(); buyer = (tngTradeBot*) _buyers.Next(); )
		{
			offers += buyer->SubmitOffers();
		}

		if (offers == 0)
		{
			break;
		}

// Tell all sellers to provisionally accept their best offers...
		for (_sellers.First(); seller = (tngTradeBot*) _sellers.Next();)
		{
			seller->AcceptOffers();
		}
	}

	for (_traders.First(); trader = (tngTradeBot*) _traders.Next(); )
	{
		trader->InactiveCheck();
	}
}

void
tngSimulation::Trade(int cycle)
{
	tngTradeBot* 		seller;

// Set the current environment cycle to be this trade cycle
	_curEnv = cycle; 

// Reset number of trades
	_trades = 0;

// Tell all sellers to initiate a PD game with their respective
// trade partners.
	for (_sellers.First(); seller = (tngTradeBot*) _sellers.Next(); )
	{
		seller->PlayPD();
	}

	if (_curGen >= (GMax() - _printGen)) 
	{
		if (_curEnv >= (IMax() - _printIt)) 
		{
			DUMP("   Trade cycle #%4d:", _curEnv+1);
			DUMP("   Total trades = %3d\n", _trades);
		}
	}
}

void
tngSimulation::AssessFitness(void)
{    
	tngTradeBot* 		bot;

	for (_traders.First(); bot = (tngTradeBot*) _traders.Next(); )
	{
		bot->CalcFitness();	
	}
	
	_pureBuyPopulation.SortByFitness();
	_pureSellPopulation.SortByFitness();
	_buysellPopulation.SortByFitness();  
}

void      
tngSimulation::DumpBots(void)
{
	// detailed info printed for the last generation
	if (_curGen == GMax()-1)
	{
		tngTradeBot* 		bot;

		for (_traders.First(); bot = (tngTradeBot*) _traders.Next(); )
		{
			bot->Dump();
		}
	}
}

void
tngSimulation::CalcScores(void)
{  
	// Determine the subpopulation and total population
	// min, max, and average fitness scores.

	// Total population values

	tngScores		total;
	float *			arrAllFit;
	if (_traderCount > 0)
	{
		arrAllFit = new float[_traderCount];
	}


	// Temporary variables
	int				i;
	tngTradeBot*	bot = 0;

	if (_pureBuyerCount > 0)
	{
		tngScores	s;

		float* arrFit = new float[_pureBuyerCount];

		for (i = 0; i < _pureBuyerCount; i++)
		{
			bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
			ASSERT_VALID(bot);

			arrFit[i] = bot->CalcFitness() ;
			arrAllFit[i] = arrFit[i];
			
			s.minScore = min(s.minScore, arrFit[i]);
			s.maxScore = max(s.maxScore, arrFit[i]);
			s.totScore += arrFit[i];
		}
		s.avgScore = s.totScore/_pureBuyerCount;
		s.stddevScore = stddevpop(arrFit, _pureBuyerCount);

		total.totScore += s.totScore;
		total.minScore = min(total.minScore, s.minScore);
		total.maxScore = max(total.maxScore, s.maxScore);

		_pbScores = s;

		delete[] arrFit;
	}

	if (_pureSellerCount > 0)
	{
		tngScores	s;

		float* arrFit = new float[_pureSellerCount];

		for (i = 0; i < _pureSellerCount; i++)
		{
			bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
			ASSERT_VALID(bot);

			arrFit[i] = bot->CalcFitness() ;
			arrAllFit[i + _pureBuyerCount] = arrFit[i];

			s.minScore = min(s.minScore, arrFit[i]);
			s.maxScore = max(s.maxScore, arrFit[i]);
			s.totScore += arrFit[i];
		}
		s.avgScore = s.totScore/_pureSellerCount;
		s.stddevScore = stddevpop(arrFit, _pureSellerCount);

		total.totScore += s.totScore;
		total.minScore = min(total.minScore, s.minScore);
		total.maxScore = max(total.maxScore, s.maxScore);

		_psScores = s;

		delete[] arrFit;
	}

	if (_buysellCount > 0)
	{
		tngScores	s;

		float* arrFit = new float[_buysellCount];

		for (i = 0; i < _buysellCount; i++)
		{
			bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
			ASSERT_VALID(bot);

			arrFit[i] = bot->CalcFitness() ;
			arrAllFit[i + _pureBuyerCount + _pureSellerCount] = arrFit[i];

			s.minScore = min(s.minScore, arrFit[i]);
			s.maxScore = max(s.maxScore, arrFit[i]);
			s.totScore += arrFit[i];
		}
		s.avgScore = s.totScore/_buysellCount;
		s.stddevScore = stddevpop(arrFit, _buysellCount);

		total.totScore += s.totScore;
		total.minScore = min(total.minScore, s.minScore);
		total.maxScore = max(total.maxScore, s.maxScore);

		_bsScores = s;

		delete[] arrFit;
	}

	if (_traderCount > 0)
	{
		total.avgScore = total.totScore/_traderCount;
		total.stddevScore = stddevpop(arrAllFit, _traderCount);

		delete[] arrAllFit;
	}

	_allScores = total;
}


void
tngSimulation::DumpInput(void)
// : ouput all the simulation parameters
{
	// first info being dumped, OVERWRITE ("w")
	FILE*	fp = _dumpFile != "" ? fopen (_dumpFile, "w") : stdout;

	if (fp == 0)
	{
		ASSERT(0);
		return;
	}
	
	struct tm *newtime;
	time_t aclock;

	time(&aclock); 
	newtime = localtime(&aclock); 
fprintf(fp, "\nCurrent Date and Time: %s\n\n", asctime(newtime));

fprintf(fp, "GMax=%d, IMax=%d\n", _gmax,  _imax );

fprintf(fp, "randSeed=%d\n", _seed );
fprintf(fp, "mutation rate=%6.3f\n", _mrate );

fprintf(fp, "TraderCount ");
fprintf(fp, "Total=%3d:  PureBuyer=%2d,    PureSeller=%2d,   BuyerSeller=%2d\n",
   _traderCount, _pureBuyerCount, _pureSellerCount, _buysellCount);

fprintf(fp, "ElitePercents:");
fprintf(fp, "          ElitePerPB=%4.2f, ElitePerPS=%4.2f, ElitePerBS=%4.2f\n",
				   _elitePercentPB, _elitePercentPS, _elitePercentBS);

fprintf(fp, "EliteCounts:  ");
fprintf(fp, "          EliteCntPB=%d,    EliteCntPS=%d,    EliteCntBS=%d\n",
				   _eliteCountPB,   _eliteCountPS,   _eliteCountBS);


fprintf(fp, "BuyerQuota=%d,  SellerQuota=%d\n", _buyerQuota, _sellerQuota);

fprintf(fp, "InitExpPayoff=%+4.2f\n", _initExpPayoff );
fprintf(fp, "ExperienceGain=%4.2f\n", _experienceGain);

fprintf(fp, "H=%+4.2f, C=%+4.2f, D=%+4.2f, L=%+4.2f, R=%+6.3f, W=%+6.3f\n",
		_temptation, _bothCoop, _bothDefect, _sucker,
		_refusalPayoff, _inactivePayoff);

fprintf(fp, "FsmStates=%d\n", _fsmStates);
fprintf(fp, "FsmMemory=%d\n\n", _fsmMemory);
			
	if (_buysellCount < 0)
	{
		_buysellCount = 0;
	}

		
	if (_dumpFile != "")
		fclose(fp);
}


void
tngSimulation::Dump(void)
{  
	// append to existing file
	FILE*	fp = _dumpFile != "" ? fopen (_dumpFile, "a+") : stdout;

	if (fp == 0)
	{
		ASSERT(0);
		return;
	}
	
// Output the subpopulation and total population
// min, max, average, and standard deviation fitness scores.

   if (_pureBuyerCount > 0)
   {
		fprintf(fp, "    PureBuyer scores...   min: %+5.4f max: %+5.4f avg: %+5.4f stdev: %+5.4f\n",
                _pbScores.minScore, _pbScores.maxScore, _pbScores.avgScore, _pbScores.stddevScore);
   }

   if (_pureSellerCount > 0)
   {
		fprintf(fp, "    PureSeller scores...  min: %+5.4f max: %+5.4f avg: %+5.4f stdev: %+5.4f\n",
                _psScores.minScore, _psScores.maxScore, _psScores.avgScore, _psScores.stddevScore);
   }

   if (_buysellCount > 0)
   {
		fprintf(fp, "    BuyerSeller scores... min: %+5.4f max: %+5.4f avg: %+5.4f stdev: %+5.4f\n",
                _bsScores.minScore, _bsScores.maxScore, _bsScores.avgScore, _bsScores.stddevScore);
   }

   fprintf(fp, "    all fitness scores... min: %+5.4f max: %+5.4f avg: %+5.4f stdev: %+5.4f\n",
                _allScores.minScore, _allScores.maxScore, _allScores.avgScore, _allScores.stddevScore);

	if (_dumpFile != "")
		fclose(fp);
}

void
tngSimulation::EvolveGen(void)
{    
	const int bKeepPhenotypes = TRUE;	// reminder that this could be a user choice if desired
	
	
	if (!bKeepPhenotypes)
	{
		ASSERT(0);

		_traders.RemoveAll();
		_buyers.RemoveAll();
		_sellers.RemoveAll();
	    
		
		// evolve genotypes and place in newly
		// created phenotypes (identity and memory are lost) 
		_pureBuyPopulation.Breed();
		_pureSellPopulation.Breed();
		_buysellPopulation.Breed();

		// rebuild trading lists and initialize bots
		InitTradeBots();
	}
	else
	{
		// evolve underlying genotypes but keep phenotypes intact, 
		// preserving identity and memory
		// (in this case the genotype can be interpreted as a set of memes)
		_pureBuyPopulation.Socialize();
		_pureSellPopulation.Socialize();
		_buysellPopulation.Socialize();  


		// since the phenotypes are not destroyed, we need to control which
		// data is reset and which is carried over to the next generation
		_PrepareBotsForNextGeneration();
	}
}

void
tngSimulation::_PrepareBotsForNextGeneration()
// : used when the phenotypes are not destroyed between generations
// : to update the FSM and internal states
{
	int				i;
	tngTradeBot*	bot = 0;

	for (i = 0; i < _pureBuyerCount; i++)
	{
		bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
		ASSERT_VALID(bot);

		bot->PrepareForNextGeneration();
	}

	for (i = 0; i < _pureSellerCount; i++)
	{
		bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		bot->PrepareForNextGeneration();
	}

	for (i = 0; i < _buysellCount; i++)
	{
		bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		bot->PrepareForNextGeneration();
	}
}

float
tngSimulation::DilemmaPayoff(int mine, int his) 
//: Returns a payoff for the prisoner's dilemma
{       
	if (mine > 0)
	{
		if (his > 0)
		{
			return(float) _bothCoop;
		}
		else
		{            
			return(float) _sucker;
		}
	}
	else              
	{
		if (his > 0)
		{
			return(float) _temptation;
		}
		else
		{            
			return(float) _bothDefect;
		}
	}
}

void
tngSimulation::MediateTrade(tngTradeBot* seller, tngTradeBot* buyer)
//: Mediate a trade between two tradebots.
{ 
	int					sellerAction;
	int					buyerAction;
	                         
	float				sellerPayoff;
	float				buyerPayoff;
		                         
	ASSERT_VALID(seller);
	ASSERT_VALID(buyer);
	
// Retrieve an action from each tradebot
	sellerAction = seller->GetAction(buyer);
	buyerAction = buyer->GetAction(seller);
	                                        
// Determine a payoff for each action	
	sellerPayoff = DilemmaPayoff(sellerAction, buyerAction);
	buyerPayoff = DilemmaPayoff(buyerAction, sellerAction);
                                         
// Tell each tradebot what happened, allowing them to update their
// respective information on each other.
	seller->Sell(buyer, sellerPayoff, buyerAction);
	buyer->Buy(seller, buyerPayoff, sellerAction);

/*
DUMP("\nTrade:(%s, %d, %5.4f) sold to (%s, %d, %5.4f)\n",
	seller->GetName(), sellerAction, sellerPayoff,
	buyer->GetName(), buyerAction, buyerPayoff);
*/

// Increment our trade count.
	++_trades;		
}


void
tngSimulation::OpenFile(const char* iniFile)
{  

	_gmax = GetInitInt(iniFile, "GMax", 10);
	_imax = GetInitInt(iniFile, "IMax", 10);
	_printGen = GetInitInt(iniFile, "PrintGen", 2);
	_printIt = GetInitInt(iniFile, "PrintIt", 3);

	_mrate = GetInitFloat(iniFile, "MutationRate", .005f);
	_seed  = GetInitInt(iniFile, "RandomSeed", 20);

	_fsmMemory = GetInitInt(iniFile, "FsmMemory", 1);
	_fsmStates = GetInitInt(iniFile, "FsmStates", 16);

	_traderCount     = GetInitInt(iniFile, "TraderCount", 9);
	_pureBuyerCount  = GetInitInt(iniFile, "PureBuyerCount", 3);
	_pureSellerCount = GetInitInt(iniFile, "PureSellerCount", 3);
	_buysellCount    = GetInitInt(iniFile, "BuySellCount", 3);
	
	// trader count is redundant input -- check for consistency
	ASSERT((_pureBuyerCount+_pureSellerCount+_buysellCount)==_traderCount);


	_buyerCount  = _pureBuyerCount  + _buysellCount;
	_sellerCount = _pureSellerCount + _buysellCount;

	_buyerQuota  = GetInitInt(iniFile, "BuyerQuota", 1);
	_sellerQuota = GetInitInt(iniFile, "SellerQuota", 1);

	_elitePercentPB = GetInitFloat(iniFile, "ElitePercentPB", 0.0f);
	_elitePercentPS = GetInitFloat(iniFile, "ElitePercentPS", 0.0f);
	_elitePercentBS = GetInitFloat(iniFile, "ElitePercentBS", 0.0f);

	_eliteCountPB  =  (int) (_elitePercentPB * _pureBuyerCount);
    _eliteCountPS  =  (int) (_elitePercentPS * _pureSellerCount);
    _eliteCountBS  =  (int) (_elitePercentBS * _buysellCount);

	// Initialize prisoner dilemma payoffs
	_bothCoop 	= GetInitFloat(iniFile, "BothCoop", 1.0f);
	_bothDefect = GetInitFloat(iniFile, "BothDefect", -1.0f);
	_sucker 	= GetInitFloat(iniFile, "Sucker", -2.0f);
	_temptation = GetInitFloat(iniFile, "Temptation", 2.0f);

	_initExpPayoff 	  = GetInitFloat(iniFile, "InitExpPayoff", 1.0f);
	_refusalPayoff    = GetInitFloat(iniFile, "RefusalPayoff", -1.0f);
	_inactivePayoff = GetInitFloat(iniFile, "InactivePayoff", 0.0f);
	_experienceGain = GetInitFloat(iniFile, "ExperienceGain", 0.0f);

	// test for obsolete field name in the ini file:
	// "wallflower" changed to "inactive" as of 8/9/2000  DJS
	if (CompareFloats(_inactivePayoff, 0.0f) == 0)
	{
		float wf = GetInitFloat(iniFile, "WallflowerPayoff", 0.0f);
		if (CompareFloats(wf, 0.0f) != 0) 
		{
			_inactivePayoff = wf;
		}
	}
}

void
tngSimulation::SaveFile(const char* iniFile)
{
	// overwrites existing file
	
	FILE*	fp = fopen (iniFile, "w");

	if (fp == 0)
	{
		return;
	}
	

    fprintf(fp, "Configuration file for the basic TNG parameters\n");
	
	time_t ltime;
	time(&ltime);
	fprintf(fp, "%s\n", ctime(&ltime));
	fprintf(fp, "[tng] \n\n");

	// write the parameters
	fprintf(fp, "GMax = %d\n", _gmax);
	fprintf(fp, "IMax = %d\n", _imax);
	fprintf(fp, "PrintGen = %d\n", _printGen);
	fprintf(fp, "PrintIt = %d\n", _printIt);
	fprintf(fp, "RandomSeed = %d\n", _seed);
	fprintf(fp, "MutationRate = %f\n", _mrate);
	fprintf(fp, "FsmStates = %d\n", _fsmStates);
	fprintf(fp, "FsmMemory = %d\n", _fsmMemory);
	
	fprintf(fp, "TraderCount = %d\n", _pureBuyerCount + _pureSellerCount + _buysellCount);
	fprintf(fp, "PureBuyerCount = %d\n", _pureBuyerCount);
	fprintf(fp, "PureSellerCount = %d\n", _pureSellerCount);
	fprintf(fp, "BuySellCount = %d\n", _buysellCount);
	fprintf(fp, "ElitePercentPB = %f\n", _elitePercentPB);
	fprintf(fp, "ElitePercentPS = %f\n", _elitePercentPS);
	fprintf(fp, "ElitePercentBS = %f\n", _elitePercentBS);
	fprintf(fp, "BuyerQuota = %d\n", _buyerQuota);
	fprintf(fp, "SellerQuota = %d\n", _sellerQuota);

	fprintf(fp, "InitExpPayoff = %f\n", _initExpPayoff);
	fprintf(fp, "Temptation = %f\n", _temptation);
	fprintf(fp, "BothCoop = %f\n", _bothCoop);
	fprintf(fp, "BothDefect = %f\n", _bothDefect);
	fprintf(fp, "Sucker = %f\n", _sucker);
	fprintf(fp, "RefusalPayoff = %f\n", _refusalPayoff);
	fprintf(fp, "InactivePayoff = %f\n", _inactivePayoff);
	fprintf(fp, "ExperienceGain = %f\n", _experienceGain);


	// documentation
    fprintf(fp, "\n\n");
    fprintf(fp, "// \n");
    fprintf(fp, "// Constraints on the user-supplied parameter values in tng.ini:\n");
    fprintf(fp, "//	\n");
    fprintf(fp, "// GMax and IMax can be any positive integers.  PrintGen can be any\n");
    fprintf(fp, "// nonnegative integer not exceeding GMax.  PrintIt can be any\n");
    fprintf(fp, "// nonnegative integer not exceeding IMax.  RandomSeed can be any\n");
    fprintf(fp, "// unsigned integer.  MutationRate can be any nonnegative number\n");
    fprintf(fp, "// less than 1.  FsmStates can be any positive integer, FsmMemory\n");
    fprintf(fp, "// can be any positive integer not exceeding 32, and TraderCount\n");
    fprintf(fp, "// can be any integer greater than 1. [Caution: Hardware-specific\n");
    fprintf(fp, "// memory limitations may further restrict FsmStates, FsmMemory,\n");
    fprintf(fp, "// and TraderCount in practice].\n");
    fprintf(fp, "//      PureBuyerCount, PureSellerCount, and BuySellCount can\n");
    fprintf(fp, "// be any nonnegative integers that exactly add up to TraderCount.\n");
    fprintf(fp, "// The percentages ElitePercentPB, ElitePercentPS, and ElitePercentBS \n");
    fprintf(fp, "// for elite pure buyers, elite pure sellers, and elite buyer-sellers \n");
    fprintf(fp, "// can be any real numbers from zero to one inclusive.  BuyerQuota can\n");
    fprintf(fp, "// be any nonnegative integer not exceeding SellerCount, and SellerQuota\n");
    fprintf(fp, "// can be any nonnegative integer.  Finally, InitExpPayoff, Temptation,\n");
    fprintf(fp, "// BothCoop, BothDefect, Sucker, RefusalPayoff, and InactivePayoff\n");
    fprintf(fp, "// can be any real numbers. ExperienceGain, for updating the inital\n");
	fprintf(fp, "// expected payoffs, can be any real numbers from zero to one inclusive.\n");
    fprintf(fp, "// \n");

		
	fclose(fp);
}


