// Simulation.cpp : Implementation of CSimulation
#include "stdafx.h"
#include "Tngcom.h"
#include "Simulation.h"
#include "Population.h"
#include "Scores.h"
   		
tngSimulation2::tngSimulation2()
{
	_gmax = 10;
	_imax = 10;

	_seed = 20;
	_mrate = .005f;
	_fsmMemory = 1;
	_fsmStates = 16;

	_traderCount     = 9;
	_pureBuyerCount  = 3;
	_pureSellerCount = 3;
	_buysellCount    = 3;

	_buyerQuota  = 1;
	_sellerQuota = 1;

	_elitePercentPB = 0.0f;
	_elitePercentPS = 0.0f;
	_elitePercentBS = 0.0f;

	_initExpPayoff 	  = +1.0f;
	_refusalPayoff    = -1.0f;
	_wallflowerPayoff = +0.0f;

	// Initialize prisoner dilemma payoffs
	_bothCoop 	= +1.0f;
	_bothDefect = -1.0f;
	_sucker 	= -2.0f;
	_temptation = +2.0f;

	_printGen = 2;
	_printIt  = 3;
}

void
tngSimulation2::Init()
{
	srand((unsigned int) _seed);

	_traderCount = _pureBuyerCount + _pureSellerCount + _buysellCount;

	// Elite counts by trader type, change introduced by LT, 6/6/98

	int   eliteCountPB  =  (int)(_elitePercentPB * _pureBuyerCount);
	int   eliteCountPS  =  (int)(_elitePercentPS * _pureSellerCount);
	int   eliteCountBS  =  (int)(_elitePercentBS * _buysellCount);
	
	if (_buysellCount < 0)
	{
		_buysellCount = 0;
	}

	_buyerCount  = _pureBuyerCount  + _buysellCount;
	_sellerCount = _pureSellerCount + _buysellCount;

	_trades = 0;

	tngTradeBot::SetMemory(_fsmMemory);
	tngTradeBot::SetStates(_fsmStates);
	

	// Assert statements changed to conform to new elite formats, LT 6/6/98
	// Note that elite can now include all members of an agent type.

    ASSERT((_pureBuyerCount  >= eliteCountPB) && (eliteCountPB >= 0));
    ASSERT((_pureSellerCount >= eliteCountPB) && (eliteCountPS >= 0));
    ASSERT((_buysellCount    >= eliteCountBS) && (eliteCountPB >= 0));

	// Changed initialization for elite sizes by trader type, LT 6/6/98
	if (_pureBuyerCount > 0)
	{
		_pureBuyPopulation.Init();
		_pureBuyPopulation.SetMutationRate(_mrate);
		_pureBuyPopulation.SetElite(eliteCountPB);
		_pureBuyPopulation.SetSize(_pureBuyerCount);
	}

	if (_pureSellerCount > 0)
	{
		_pureSellPopulation.Init();
		_pureSellPopulation.SetMutationRate(_mrate);
		_pureSellPopulation.SetElite(eliteCountPS);
		_pureSellPopulation.SetSize(_pureSellerCount);
	}

	if (_buysellCount > 0)
	{
		_buysellPopulation.Init();
		_buysellPopulation.SetMutationRate(_mrate);
		_buysellPopulation.SetElite(eliteCountBS);
		_buysellPopulation.SetSize(_buysellCount);
	}
}

void
tngSimulation2::OpenFile(LPCTSTR iniFile)
{  

	_gmax = GetInitInt(iniFile, "GMax", 10);
	_imax = GetInitInt(iniFile, "IMax", 10);

	_mrate = GetInitFloat(iniFile, "MutationRate", .005f);
	_seed  = GetInitInt(iniFile, "RandomSeed", 20);

	_fsmMemory = GetInitInt(iniFile, "FsmMemory", 1);
	_fsmStates = GetInitInt(iniFile, "FsmStates", 16);

	_pureBuyerCount  = GetInitInt(iniFile, "PureBuyerCount", 3);
	_pureSellerCount = GetInitInt(iniFile, "PureSellerCount", 3);
	_buysellCount    = GetInitInt(iniFile, "BuySellCount", 3);

	_buyerQuota  = GetInitInt(iniFile, "BuyerQuota", 1);
	_sellerQuota = GetInitInt(iniFile, "SellerQuota", 1);

	_elitePercentPB = GetInitFloat(iniFile, "ElitePercentPB", 0.0f);
	_elitePercentPS = GetInitFloat(iniFile, "ElitePercentPS", 0.0f);
	_elitePercentBS = GetInitFloat(iniFile, "ElitePercentBS", 0.0f);

	// Initialize prisoner dilemma payoffs
	_bothCoop 	= GetInitFloat(iniFile, "BothCoop", 1.0f);
	_bothDefect = GetInitFloat(iniFile, "BothDefect", -1.0f);
	_sucker 	= GetInitFloat(iniFile, "Sucker", -2.0f);
	_temptation = GetInitFloat(iniFile, "Temptation", 2.0f);

	_initExpPayoff 	  = GetInitFloat(iniFile, "InitExpPayoff", 1.0f);
	_refusalPayoff    = GetInitFloat(iniFile, "RefusalPayoff", -1.0f);
	_wallflowerPayoff = GetInitFloat(iniFile, "WallflowerPayoff", 0.0f);
}

IScores*
tngSimulation2::GetScores(tngPOPULATION ePop)
{
	IScores*				piScores = 0;
	CComObject<CScores>*	pScores = 0;
	tngScores*				p = 0;
	HRESULT					hr = 0;
 
	hr = CComObject<CScores>::CreateInstance(&pScores);
	
	if(FAILED(hr))
	{
		return 0;
	}

	switch (ePop)
	{
	case tngALL:
		p = &_allScores;
		break;

	case tngPURE_BUYER:
		p = &_pbScores;
		break;

	case tngPURE_SELLER:
		p = &_psScores;
		break;

	case tngBUYER_SELLER:
		p = &_bsScores;
		break;

	default:
		break;
	}

	pScores->put_Min(p->minScore);
	pScores->put_Max(p->maxScore);
	pScores->put_Avg(p->avgScore);
	pScores->put_Total(p->totScore);
	pScores->put_StdDev(p->stddevScore);

	piScores = pScores;
	piScores->AddRef();

	return piScores;
}

IPopulation*
tngSimulation2::GetPopulation(tngPOPULATION ePop)
{
	IPopulation*				piPopulation = 0;
	CComObject<CPopulation>*	pPopulation = 0;
	HRESULT						hr = 0;
 
	hr = CComObject<CPopulation>::CreateInstance(&pPopulation);
	
	if(FAILED(hr))
	{
		return 0;
	}

	switch (ePop)
	{
	case tngALL:
		break;

	case tngPURE_BUYER:
		pPopulation->Init(&_pureBuyPopulation);
		break;

	case tngPURE_SELLER:
		pPopulation->Init(&_pureSellPopulation);
		break;

	case tngBUYER_SELLER:
		pPopulation->Init(&_buysellPopulation);
		break;

	default:
		return 0;
	}

	piPopulation = pPopulation;
	piPopulation->AddRef();

	return piPopulation;
}

/////////////////////////////////////////////////////////////////////////////
// CSimulation

CSimulation::CSimulation()
{
}

CSimulation::~CSimulation()
{
}

STDMETHODIMP CSimulation::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_ISimulation
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (InlineIsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

STDMETHODIMP CSimulation::get_GMax(short *pVal)
{
	(*pVal) = _sim.GMax();

	return S_OK;
}

STDMETHODIMP CSimulation::put_GMax(short newVal)
{
	_sim.SetGMax(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_IMax(short *pVal)
{
	(*pVal) = _sim.IMax();

	return S_OK;
}

STDMETHODIMP CSimulation::put_IMax(short newVal)
{
	_sim.SetIMax(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_RandomSeed(short *pVal)
{
	(*pVal) = _sim.GetRandomSeed();

	return S_OK;
}

STDMETHODIMP CSimulation::put_RandomSeed(short newVal)
{
	_sim.SetRandomSeed(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_MutationRate(float *pVal)
{
	(*pVal) = _sim.GetMutationRate();

	return S_OK;
}

STDMETHODIMP CSimulation::put_MutationRate(float newVal)
{
	_sim.SetMutationRate(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_FsmStates(short *pVal)
{
	(*pVal) = _sim.GetFsmStates();

	return S_OK;
}

STDMETHODIMP CSimulation::put_FsmStates(short newVal)
{
	_sim.SetFsmStates(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_FsmMemory(short *pVal)
{
	(*pVal) = _sim.GetFsmMemory();

	return S_OK;
}

STDMETHODIMP CSimulation::put_FsmMemory(short newVal)
{
	_sim.SetFsmMemory(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::GetScores(tngPOPULATION ePop, IScores** pVal)
{
	(*pVal) = _sim.GetScores(ePop);

	return S_OK;
}

STDMETHODIMP CSimulation::GetPopulation(tngPOPULATION ePop, IPopulation** pVal)
{
	(*pVal) = _sim.GetPopulation(ePop);

	return S_OK;
}

STDMETHODIMP CSimulation::get_PureBuyerCount(short *pVal)
{
	(*pVal) = _sim.GetPureBuyerCount();

	return S_OK;
}

STDMETHODIMP CSimulation::put_PureBuyerCount(short newVal)
{
	_sim.SetPureBuyerCount(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_PureSellerCount(short *pVal)
{
	(*pVal) = _sim.GetPureSellerCount();

	return S_OK;
}

STDMETHODIMP CSimulation::put_PureSellerCount(short newVal)
{
	_sim.SetPureSellerCount(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_BuysellCount(short *pVal)
{
	(*pVal) = _sim.GetBuysellCount();

	return S_OK;
}

STDMETHODIMP CSimulation::put_BuysellCount(short newVal)
{
	_sim.SetBuysellCount(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_Payoff(tngPAYOFF ePayoff, float *pVal)
{
	switch (ePayoff)
	{
	case tngBOTH_COOP:
		(*pVal) = _sim.GetCoopPayoff();
		break;

	case tngBOTH_DEFECT:
		(*pVal) = _sim.GetDefectPayoff();
		break;

	case tngTEMPTATION:
		(*pVal) = _sim.GetTemptationPayoff();
		break;

	case tngSUCKER:
		(*pVal) = _sim.GetSuckerPayoff();
		break;

	case tngINIT_EXP:
		(*pVal) = _sim.GetInitExpPayoff();
		break;

	case tngWALLFLOWER:
		(*pVal) = _sim.GetWallflowerPayoff();
		break;

	case tngREFUSAL:
		(*pVal) = _sim.GetRefusalPayoff();
		break;

	default:
		return E_INVALIDARG;
	}

	return S_OK;
}

STDMETHODIMP CSimulation::put_Payoff(tngPAYOFF ePayoff, float newVal)
{
	switch (ePayoff)
	{
	case tngBOTH_COOP:
		_sim.SetCoopPayoff(newVal);
		break;

	case tngBOTH_DEFECT:
		_sim.SetDefectPayoff(newVal);
		break;

	case tngTEMPTATION:
		_sim.SetTemptationPayoff(newVal);
		break;

	case tngSUCKER:
		_sim.SetSuckerPayoff(newVal);
		break;

	case tngINIT_EXP:
		_sim.SetInitExpPayoff(newVal);
		break;

	case tngWALLFLOWER:
		_sim.SetWallflowerPayoff(newVal);
		break;

	case tngREFUSAL:
		_sim.SetRefusalPayoff(newVal);
		break;

	default:
		return E_INVALIDARG;
	}

	return S_OK;
}

STDMETHODIMP CSimulation::get_ElitePercent(tngPOPULATION ePop, float *pVal)
{
	switch (ePop)
	{
	case tngPURE_BUYER:
		(*pVal) = _sim.GetElitePercentPB();
		break;

	case tngPURE_SELLER:
		(*pVal) = _sim.GetElitePercentPS();
		break;

	case tngBUYER_SELLER:
		(*pVal) = _sim.GetElitePercentBS();
		break;

	default:
		return E_INVALIDARG;
		break;
	}

	return S_OK;
}

STDMETHODIMP CSimulation::put_ElitePercent(tngPOPULATION ePop, float newVal)
{
	switch (ePop)
	{
	case tngPURE_BUYER:
		_sim.SetElitePercentPB(newVal);
		break;

	case tngPURE_SELLER:
		_sim.SetElitePercentPS(newVal);
		break;

	case tngBUYER_SELLER:
		_sim.SetElitePercentBS(newVal);
		break;

	default:
		return E_INVALIDARG;
		break;
	}

	return S_OK;
}

STDMETHODIMP CSimulation::get_BuyerQuota(short *pVal)
{
	(*pVal) = _sim.GetBuyerQuota();

	return S_OK;
}

STDMETHODIMP CSimulation::put_BuyerQuota(short newVal)
{
	_sim.SetBuyerQuota(newVal);

	return S_OK;
}

STDMETHODIMP CSimulation::get_SellerQuota(short *pVal)
{
	(*pVal) = _sim.GetSellerQuota();

	return S_OK;
}

STDMETHODIMP CSimulation::put_SellerQuota(short newVal)
{
	_sim.SetSellerQuota(newVal);

	return S_OK;
}


STDMETHODIMP CSimulation::Go()
{
	int            g, i;

	_sim.Init();

	// Start the generation cycle loop.

	for (g = 0; g < _sim.GMax(); g++)
	{
		if (g > 0)
		{
			_sim.EvolveGen();                 // evolution step
		}

		_sim.InitGen(g);

		Fire_GenerationBegin(g);

		for (i = 0; i < _sim.IMax(); i++)    // trade cycle loop
		{
			Fire_TradeCycleBegin(i);

			_sim.MatchTraders();
			_sim.Trade(i);
			
			Fire_TradeCycleEnd(i);
		}
											// environmental step
		_sim.AssessFitness();
		_sim.CalcScores();		

		Fire_GenerationEnd(g);
	}

	return S_OK;
}


STDMETHODIMP CSimulation::Run()
{
	_bPaused = false;
	_bStopped = false;
	_nGen = 0;

	_sim.Init();

	_Run();

	return S_OK;
}

STDMETHODIMP CSimulation::Pause()
{
	_bPaused = true;

	return S_OK;
}

STDMETHODIMP CSimulation::Resume()
{
	_bPaused = false;

	_Run();

	return S_OK;
}

STDMETHODIMP CSimulation::Stop()
{
	_bStopped = true;

	// if the simulation is already paused we need to fire the event
	// immediately, otherwise it will wait until the end of the current
	// generation....see _Run()
	if (_bPaused)
	{
		Fire_SimStopped();
	}

	return S_OK;
}


void CSimulation::_Run()
{
	int			i;

	Fire_SimRunning();

	// Continue the generation cycle loop.
	for ( ; _nGen < _sim.GMax(); _nGen++)
	{
		if (_nGen > 0)
		{
			_sim.EvolveGen();				// evolution step
		}

		_sim.InitGen(_nGen);

		Fire_GenerationBegin(_nGen);

		for (i = 0; i < _sim.IMax(); i++)	// trade cycle loop
		{
			Fire_TradeCycleBegin(i);

			_sim.MatchTraders();
			_sim.Trade(i);
			
			Fire_TradeCycleEnd(i);
		}
											// environmental step
		_sim.AssessFitness();
		_sim.CalcScores();		

		Fire_GenerationEnd(_nGen);

		if (_bPaused)
		{
			// increment the generation count so it is set properly
			// if and when the simulation is resumed
			_nGen++;

			Fire_SimPaused();
			return;
		}

		if (_bStopped)
		{
			Fire_SimStopped();
			return;
		}
	}

	Fire_SimFinished();

	return;
}

STDMETHODIMP CSimulation::OpenFile(BSTR inifile)
{	
	USES_CONVERSION;

	_sim.OpenFile(OLE2CT(inifile));

	return S_OK;
}
