//
// tngSim.cpp (for tnglab.101 -- but not for use with tngshell)
//
// Last Revised: 29 June 1998
//

#include <stdio.h>
#include <time.h>
#include "tngsim.h"

/////////////////////////////////////////////////////////
// population standard deviation utility
float stddevpop(float * val, int count)
{
	if (count <= 0 || val == 0) 
		return 0.0f;

	int i;
	float sum = 0.0;
	for (i = 0; i < count; i++)
	{
		sum += val[i];
	}

	float avg = sum / (float) count;	

	float diff = 0.0;
	float variance = 0.0;
	for (i = 0; i < count; i++)
	{
		diff = val[i] - avg;
		variance += diff * diff;
	}
	variance /= count;	// note "sample standard deviation" would divide by (count - 1) instead

	float stdDev = (float) sqrt(variance);

	return stdDev;
}

tngScores::tngScores()
{
	minScore = +1000.0f;
	avgScore = 0.0f;
	maxScore = -1000.0f;
	totScore = 0.0f;
	stddevScore = 0.0f;
}


//////////////////////////
//
// tngSimulation methods

tngSimulation::tngSimulation(void)
{
	_gmax = 0;
	_imax = 0;

	_seed = 0;
	_mrate = 0.0;
	_fsmStates = 0;
	_fsmMemory = 0;

	_elitePercentPB = 0.0;
	_elitePercentPS = 0.0;
	_elitePercentBS = 0.0;


	_buyerQuota = 0;     
	_sellerQuota = 0;    
	_initExpPayoff = 0.0;
	_refusalPayoff = 0.0;  
	_wallflowerPayoff = 0.0; 
	_bothCoop = 0.0;         
	_bothDefect = 0.0;       
	_temptation = 0.0;       
	_sucker = 0.0;           
	_traderCount = 0;      
	_buyerCount = 0;       
	_sellerCount = 0;      
	_pureBuyerCount = 0;   
	_pureSellerCount = 0;  
	_buysellCount = 0;     
	_trades = 0;           
	_printGen = 0;         
	_printIt = 0;          
}

tngSimulation::~tngSimulation(void)
{ 
	_traders.RemoveAll();
	_buyers.RemoveAll();
	_sellers.RemoveAll();
}

void
tngSimulation::Init(void)
{    		
	char*	iniFile = "tng.ini";
	_mrate = GetInitFloat(iniFile, "MutationRate", .005f);

// Method of elite specification changed by LT, 6/6/98

	_elitePercentPB = GetInitFloat(iniFile, "ElitePercentPB", 0.0f);
	_elitePercentPS = GetInitFloat(iniFile, "ElitePercentPS", 0.0f);
	_elitePercentBS = GetInitFloat(iniFile, "ElitePercentBS", 0.0f);

	_seed  = GetInitInt(iniFile, "RandomSeed", 20);

	_fsmMemory = GetInitInt(iniFile, "FsmMemory", 1);
	_fsmStates = GetInitInt(iniFile, "FsmStates", 16);

	srand((unsigned int) _seed);

	_gmax = GetInitInt(iniFile, "GMax", 10);
	_imax = GetInitInt(iniFile, "IMax", 10);

	_traderCount     = GetInitInt(iniFile, "TraderCount", 9);
	_pureBuyerCount  = GetInitInt(iniFile, "PureBuyerCount", 3);
	_pureSellerCount = GetInitInt(iniFile, "PureSellerCount", 3);
	_buysellCount    = GetInitInt(iniFile, "BuySellCount", 3);

	ASSERT((_pureBuyerCount+_pureSellerCount+_buysellCount)==_traderCount);

// Elite counts by trader type, change introduced by LT, 6/6/98

	int   eliteCountPB  =  (int) (_elitePercentPB * _pureBuyerCount);
    int   eliteCountPS  =  (int) (_elitePercentPS * _pureSellerCount);
    int   eliteCountBS  =  (int) (_elitePercentBS * _buysellCount);

	_buyerCount  = _pureBuyerCount  + _buysellCount;
	_sellerCount = _pureSellerCount + _buysellCount;

	_buyerQuota  = GetInitInt(iniFile, "BuyerQuota", 1);
	_sellerQuota = GetInitInt(iniFile, "SellerQuota", 1);

	_initExpPayoff 	  = GetInitFloat(iniFile, "InitExpPayoff", 1.0f);
	_refusalPayoff    = GetInitFloat(iniFile, "RefusalPayoff", -1.0f);
	_wallflowerPayoff = GetInitFloat(iniFile, "WallflowerPayoff", 0.0f);

// Initialize prisoner dilemma payoffs

	_bothCoop 	= GetInitFloat(iniFile, "BothCoop", 1.0f);
	_bothDefect = GetInitFloat(iniFile, "BothDefect", -1.0f);
	_sucker 	= GetInitFloat(iniFile, "Sucker", -2.0f);
	_temptation = GetInitFloat(iniFile, "Temptation", 2.0f);

	_printGen = GetInitInt(iniFile, "PrintGen", 2);
	_printIt  = GetInitInt(iniFile, "PrintIt", 3);

	_trades = 0;

	tngTradeBot::SetMemory(_fsmMemory);
	tngTradeBot::SetStates(_fsmStates);

	struct tm *newtime;
	time_t aclock;

	time(&aclock); 
	newtime = localtime(&aclock); 
printf("\nCurrent Date and Time: %s\n\n", asctime(newtime));

printf("GMax=%d, IMax=%d\n", _gmax,  _imax );

printf("randSeed=%d\n", _seed );
printf("mutation rate=%6.3f\n", _mrate );

// Changed printf statements to handle changed elite format, LT 6/7/98

printf("TraderCount ");
printf("Total=%3d:  PureBuyer=%2d,    PureSeller=%2d,   BuyerSeller=%2d\n",
   _traderCount, _pureBuyerCount, _pureSellerCount, _buysellCount);

printf("ElitePercents:");
printf("          ElitePerPB=%4.2f, ElitePerPS=%4.2f, ElitePerBS=%4.2f\n",
				   _elitePercentPB, _elitePercentPS, _elitePercentBS);

printf("EliteCounts:  ");
printf("          EliteCntPB=%d,    EliteCntPS=%d,    EliteCntBS=%d\n",
				   eliteCountPB,   eliteCountPS,   eliteCountBS);


printf("BuyerQuota=%d,  SellerQuota=%d\n", _buyerQuota, _sellerQuota);

printf("InitExpPayoff=%+4.2f\n", _initExpPayoff );

printf("H=%+4.2f, C=%+4.2f, D=%+4.2f, L=%+4.2f, R=%+6.3f, W=%+6.3f\n",
		_temptation, _bothCoop, _bothDefect, _sucker,
		_refusalPayoff, _wallflowerPayoff);

printf("FsmStates=%d\n", _fsmStates);
printf("FsmMemory=%d\n\n", _fsmMemory);
			
	if (_buysellCount < 0)
	{
		_buysellCount = 0;
	}

// Assert statements changed to conform to new elite formats, LT 6/6/98
// Note that elite can now include all members of an agent type.

	ASSERT((_pureBuyerCount  >= eliteCountPB) && (eliteCountPB >= 0));
	ASSERT((_pureSellerCount >= eliteCountPS) && (eliteCountPS >= 0));
	ASSERT((_buysellCount    >= eliteCountBS) && (eliteCountBS >= 0));

// Changed initialization for elite sizes by trader type, LT 6/6/98

	if (_pureBuyerCount > 0)
	{
		_pureBuyPopulation.Init();
		_pureBuyPopulation.SetMutationRate(_mrate);
		_pureBuyPopulation.SetElite(eliteCountPB);
		_pureBuyPopulation.SetSize(_pureBuyerCount);
	}

	if (_pureSellerCount > 0)
	{
		_pureSellPopulation.Init();
		_pureSellPopulation.SetMutationRate(_mrate);
		_pureSellPopulation.SetElite(eliteCountPS);
		_pureSellPopulation.SetSize(_pureSellerCount);
	}

	if (_buysellCount > 0)
	{
		_buysellPopulation.Init();
		_buysellPopulation.SetMutationRate(_mrate);
		_buysellPopulation.SetElite(eliteCountBS);
		_buysellPopulation.SetSize(_buysellCount);
	}
}

void
tngSimulation::InitGen(int g)
{
	int		i;
	tngTradeBot*		bot = 0;

	ASSERT(_traders.IsEmpty());
	ASSERT(_buyers.IsEmpty());
	ASSERT(_sellers.IsEmpty());

	_curGen = g;
	printf ("\nGeneration #%3d\n", _curGen+1);

	for (i = 0; i < _pureBuyerCount; i++)
	{
		bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _pureSellerCount; i++)
	{
		bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _buysellCount; i++)
	{
		bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_traders.Append(bot);
	}

	for (i = 0; i < _pureBuyerCount; i++)
	{
		bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_buyers.Append(bot);

		bot->Init(this, _traders, _buyerQuota, 0);
	}

	for (i = 0; i < _pureSellerCount; i++)
	{
		bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_sellers.Append(bot);

		bot->Init(this, _traders, 0, _sellerQuota);
	}

	for (i = 0; i < _buysellCount; i++)
	{
		bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		_buyers.Append(bot);
		_sellers.Append(bot);

		bot->Init(this, _traders, _buyerQuota, _sellerQuota);
	}
}

void
tngSimulation::MatchTraders(void) 
{
	tngTradeBot* 		buyer;
	tngTradeBot* 		seller;
	tngTradeBot* 		trader;
	int					offers;

	for (_buyers.First(); buyer = (tngTradeBot*) _buyers.Next(); )
	{
		buyer->PrepareOffers();
	}
                   
// Tell all buyers to submit offers and all sellers to
// provisionally accept offers until the buyers stop making
// offers.
	while(1)
	{
		offers = 0;

// Tell all buyers to submit their offers...
		for (_buyers.First(); buyer = (tngTradeBot*) _buyers.Next(); )
		{
			offers += buyer->SubmitOffers();
		}

		if (offers == 0)
		{
			break;
		}

// Tell all sellers to provisionally accept their best offers...
		for (_sellers.First(); seller = (tngTradeBot*) _sellers.Next();)
		{
			seller->AcceptOffers();
		}
	}

	for (_traders.First(); trader = (tngTradeBot*) _traders.Next(); )
	{
		trader->WallflowerCheck();
	}
}

void
tngSimulation::Trade(int cycle)
{
	tngTradeBot* 		seller;

// Set the current environment cycle to be this trade cycle
	_curEnv = cycle; 

// Reset number of trades
	_trades = 0;

// Tell all sellers to initiate a PD game with their respective
// trade partners.
	for (_sellers.First(); seller = (tngTradeBot*) _sellers.Next(); )
	{
		seller->PlayPD();
	}

	if (_curGen >= (GMax() - _printGen)) 
	{
		if (_curEnv >= (IMax() - _printIt)) 
		{
			printf ("   Trade cycle #%4d:", _curEnv+1);
			printf ("   Total trades = %3d\n", _trades);
		}
	}
}

void
tngSimulation::AssessFitness(void)
{    
	tngTradeBot* 		bot;

	for (_traders.First(); bot = (tngTradeBot*) _traders.Next(); )
	{
		bot->CalcFitness();	

		if (_curGen == GMax()-1)
		{
			bot->Dump();
		}
	}
	
	_pureBuyPopulation.SortByFitness();
	_pureSellPopulation.SortByFitness();
	_buysellPopulation.SortByFitness();  
}

void
tngSimulation::CalcScores(void)
{  
	// Determine the subpopulation and total population
	// min, max, and average fitness scores.

	// Total population values

	tngScores		total;
	float *			arrAllFit;
	if (_traderCount > 0)
	{
		arrAllFit = new float[_traderCount];
	}


	// Temporary variables
	int				i;
	tngTradeBot*	bot = 0;

	if (_pureBuyerCount > 0)
	{
		tngScores	s;

		float* arrFit = new float[_pureBuyerCount];

		for (i = 0; i < _pureBuyerCount; i++)
		{
			bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
			ASSERT_VALID(bot);

			arrFit[i] = bot->CalcFitness() ;
			arrAllFit[i] = arrFit[i];
			
			s.minScore = min(s.minScore, arrFit[i]);
			s.maxScore = max(s.maxScore, arrFit[i]);
			s.totScore += arrFit[i];
		}
		s.avgScore = s.totScore/_pureBuyerCount;
		s.stddevScore = stddevpop(arrFit, _pureBuyerCount);

		total.totScore += s.totScore;
		total.minScore = min(total.minScore, s.minScore);
		total.maxScore = max(total.maxScore, s.maxScore);

		_pbScores = s;

		delete[] arrFit;
	}

	if (_pureSellerCount > 0)
	{
		tngScores	s;

		float* arrFit = new float[_pureSellerCount];

		for (i = 0; i < _pureSellerCount; i++)
		{
			bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
			ASSERT_VALID(bot);

			arrFit[i] = bot->CalcFitness() ;
			arrAllFit[i + _pureBuyerCount] = arrFit[i];

			s.minScore = min(s.minScore, arrFit[i]);
			s.maxScore = max(s.maxScore, arrFit[i]);
			s.totScore += arrFit[i];
		}
		s.avgScore = s.totScore/_pureSellerCount;
		s.stddevScore = stddevpop(arrFit, _pureSellerCount);

		total.totScore += s.totScore;
		total.minScore = min(total.minScore, s.minScore);
		total.maxScore = max(total.maxScore, s.maxScore);

		_psScores = s;

		delete[] arrFit;
	}

	if (_buysellCount > 0)
	{
		tngScores	s;

		float* arrFit = new float[_buysellCount];

		for (i = 0; i < _buysellCount; i++)
		{
			bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
			ASSERT_VALID(bot);

			arrFit[i] = bot->CalcFitness() ;
			arrAllFit[i + _pureBuyerCount + _pureSellerCount] = arrFit[i];

			s.minScore = min(s.minScore, arrFit[i]);
			s.maxScore = max(s.maxScore, arrFit[i]);
			s.totScore += arrFit[i];
		}
		s.avgScore = s.totScore/_buysellCount;
		s.stddevScore = stddevpop(arrFit, _buysellCount);

		total.totScore += s.totScore;
		total.minScore = min(total.minScore, s.minScore);
		total.maxScore = max(total.maxScore, s.maxScore);

		_bsScores = s;

		delete[] arrFit;
	}

	if (_traderCount > 0)
	{
		total.avgScore = total.totScore/_traderCount;
		total.stddevScore = stddevpop(arrAllFit, _traderCount);

		delete[] arrAllFit;
	}

	_allScores = total;
}

void
tngSimulation::Dump(void)
{  
// Output the subpopulation and total population
// min, max, average, and standard deviation fitness scores.

   CalcScores();

   if (_pureBuyerCount > 0)
   {
		printf("    PureBuyer scores...   min: %+5.4f max: %+5.4f avg: %+5.4f stdev: %+5.4f\n",
                _pbScores.minScore, _pbScores.maxScore, _pbScores.avgScore, _pbScores.stddevScore);
   }

   if (_pureSellerCount > 0)
   {
		printf("    PureSeller scores...  min: %+5.4f max: %+5.4f avg: %+5.4f stdev: %+5.4f\n",
                _psScores.minScore, _psScores.maxScore, _psScores.avgScore, _psScores.stddevScore);
   }

   if (_buysellCount > 0)
   {
		printf("    BuyerSeller scores... min: %+5.4f max: %+5.4f avg: %+5.4f stdev: %+5.4f\n",
                _bsScores.minScore, _bsScores.maxScore, _bsScores.avgScore, _bsScores.stddevScore);
   }

   printf("    all fitness scores... min: %+5.4f max: %+5.4f avg: %+5.4f stdev: %+5.4f\n",
                _allScores.minScore, _allScores.maxScore, _allScores.avgScore, _allScores.stddevScore);
}

void
tngSimulation::EvolveGen(void)
{    
	_traders.RemoveAll();
	_buyers.RemoveAll();
	_sellers.RemoveAll();
	             
	const int bKeepPhenotypes = TRUE;	// reminder that this could be a user choice if desired
	if (!bKeepPhenotypes)
	{
		// evolve genotypes and place in newly
		// created phenotypes (identity and memory are lost) 
		_pureBuyPopulation.Breed();
		_pureSellPopulation.Breed();
		_buysellPopulation.Breed();
	}
	else
	{
		// evolve underlying genotypes but keep phenotypes intact, 
		// preserving identity and memory
		// (in this case the genotype can be interpreted as a set of memes)
		_pureBuyPopulation.SocialLearn();
		_pureSellPopulation.SocialLearn();
		_buysellPopulation.SocialLearn();  


		// need to set each bots total payoff to zero because of the way the fitness
		// calculation is done (total payoff / trades per cycle)
		_ResetTotalPayoffs();
	}
}

void
tngSimulation::_ResetTotalPayoffs()
{
	int				i;
	tngTradeBot*	bot = 0;

	for (i = 0; i < _pureBuyerCount; i++)
	{
		bot = (tngTradeBot*) _pureBuyPopulation.GetPType(i);
		ASSERT_VALID(bot);

		bot->ResetTotalPayoff() ;
	}

	for (i = 0; i < _pureSellerCount; i++)
	{
		bot = (tngTradeBot*) _pureSellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		bot->ResetTotalPayoff() ;
	}

	for (i = 0; i < _buysellCount; i++)
	{
		bot = (tngTradeBot*) _buysellPopulation.GetPType(i);
		ASSERT_VALID(bot);

		bot->ResetTotalPayoff() ;
	}
}

float
tngSimulation::DilemmaPayoff(int mine, int his) 
//: Returns a payoff for the prisoner's dilemma
{       
	if (mine > 0)
	{
		if (his > 0)
		{
			return(float) _bothCoop;
		}
		else
		{            
			return(float) _sucker;
		}
	}
	else              
	{
		if (his > 0)
		{
			return(float) _temptation;
		}
		else
		{            
			return(float) _bothDefect;
		}
	}
}

void
tngSimulation::MediateTrade(tngTradeBot* seller, tngTradeBot* buyer)
//: Mediate a trade between two tradebots.
{ 
	int					sellerAction;
	int					buyerAction;
	                         
	float				sellerPayoff;
	float				buyerPayoff;
		                         
	ASSERT_VALID(seller);
	ASSERT_VALID(buyer);
	
// Retrieve an action from each tradebot
	sellerAction = seller->GetAction(buyer);
	buyerAction = buyer->GetAction(seller);
	                                        
// Determine a payoff for each action	
	sellerPayoff = DilemmaPayoff(sellerAction, buyerAction);
	buyerPayoff = DilemmaPayoff(buyerAction, sellerAction);
                                         
// Tell each tradebot what happened, allowing them to update their
// respective information on each other.
	seller->Sell(buyer, sellerPayoff, buyerAction);
	buyer->Buy(seller, buyerPayoff, sellerAction);

/*
printf("\nTrade:(%s, %d, %5.4f) sold to (%s, %d, %5.4f)\n",
	seller->GetName(), sellerAction, sellerPayoff,
	buyer->GetName(), buyerAction, buyerPayoff);
*/

// Increment our trade count.
	++_trades;		
}













